#!/usr/bin/perl -w
#
# $Id: mbox2imap.pl,v 1.8 2004/08/29 11:28:13 norgaard Exp $
#
# This script migrates unix-style mailbox to (cyrus) imap server. It is
# designed to migrate ALL user and system mailboxes. It uses plaintext
# authentication so while migrating have that enabled.
#
use strict;
use warnings;
#
use File::Basename;
use File::Copy;
use Getopt::Std;                # Command line options
use IMAP::Admin;                # To create mailbox
use Mail::IMAPClient;           # To migrate mailbox

$| = 1;

my ($b,$f,$m,$s,$t) = (0,0,0,0,0);
my ($box, @users, %opts);
# Initializing:

getopts('hvi:p:q:s:u:',\%opts);

if ($opts{h}) { # Help text
  print <<'END_OF_HELP';
Usage: mbox2imap.pl [-u <admin> -s <server> -p <passwd> <user> ... ]
See 'perldoc mbox2imap.pl' for more details.
END_OF_HELP
exit 0;
}

my $debug  = $opts{'v'};
my $glob   = $opts{'g'};                  # Argumenst are fileglobs, not users
my $passwd = $opts{'p'} ? $opts{'p'} : '';
my $quota  = $opts{'q'} ? $opts{'q'} : 0; # In kilobytes!
my $server = $opts{'s'} ? $opts{'s'} : 'localhost';
my $admin  = $opts{'u'} ? $opts{'u'} : 'cyrus';
my $tmpdir = '/tmp';

if (!$passwd) {
  print "Enter password for $admin: ";
  $passwd = <STDIN>;
  chomp $passwd;
}

if ($glob) { # Arguments are file globs
  foreach my $glob (@ARGV) {
    foreach my $mbox (glob($glob)) {
      next if basename($mbox) =~ /^\./o;
      $box->{$mbox} = basename($mbox);
    };
  };
}
else {
  if (@ARGV == 1 and $ARGV[0] =~ /^\@(.*)$/) {
    @users = getusers($ARGV[0]);
  }
  elsif (@ARGV) {
    @users = @ARGV;
  }
  else {
    @users = getusers();
  }
  $box = getmbox(@users);
}

$box or die "No mailboxes to transfer.\n";

#####################################################################
#
# Connect to IMAP server as admin

my $imap = IMAP::Admin->new(
    'Server'   => $server,
    'Login'    => $admin,
    'Password' => $passwd,
    'CRAM'     => 2,
    'Debug'    => $debug)
  or die "Could not authenticate: $@\n";

#####################################################################
#
# Administrative part: Add accounts, set permissions

foreach my $user (@users) {
  print "Checking account for $user: ";
  if ($imap->list("user.$user")) {
    print "found.\n";
  }
  else {
    print "not found.\n", "Creating account for $user: ";
    $imap->create("user.$user")
      and ( warn "Failed: ", $imap->error(), "\n" and next );
    print "OK\n";
  };

  print "Setting permissions: "; # All permissions!
  my $sub = "user.$user";
  $sub =~ s/@/.*@/;
  foreach my $folder ("user.$user", $imap->list($sub)) {
    print "Setting permissions on $folder: " if $debug;
    $imap->set_acl("$folder", $user, "lrswipcda", $admin, "lrswipcda") and
      (warn "failed setting permissions on $folder: $@\n" and next);
    print "OK\n" if $debug;
  }
  print "OK\n";

  if ($quota) {  # Set quota on top folder
    $imap->set_quota("user.$user",$quota);
  };
};

#####################################################################
#
# Transfer mail

my $start = time();

my $client = Mail::IMAPClient->new(
    'Server'       => $server,
    'User'         => $admin,
    'Password'     => $passwd,
    'Debug'        => $debug )
  or die "Failed connect to $server. $@";

foreach my $mbox (sort keys %{$box}) {
  next if (-d $mbox and -z $mbox);

  print "Transfering messages from $mbox to $box->{$mbox}\n";
  my ($h,$i,$j,$k) = mbox2imap($client,$mbox,$box->{$mbox});
  $s += -s $mbox;
  $b += $h;
  $f += $i;
  $m += $j;
  $t += $k;
};

# Disconnect: DONE!
$client->disconnect;

print "Finnished processing files. Summary:\n";

print "Transfered $t mails/$b bytes of $m total/$s bytes.\n";
print "Failed transfering $f mails, ", 100*$f/$t, "% errors\n";
print "Processed: ", $m/(time()-$start), " mails/second, ", $s/(time()-$start), " bytes/second\n";
print "Transfered:", $t/(time()-$start), " mails/second, ", $b/(time()-$start), " bytes/second\n";
print "Time: ", time()-$start, " seconds\n";

#####################################################################
#
# Admin: Revoke privileges for admin user

print "Revoking admin's permissions: "; # All permissions!
foreach my $folder (values %{ $box }) {
  $imap->delete_acl("$folder",$admin) and
    warn "failed revoke on $folder: $@\n" and next;
}
print "OK\n";

$imap->close;

#####################################################################
#
# Subrutines

sub getusers {
  # This rutine parses the /etc/passwd file and returns a list of real users.
  # NOTE: users with "nologin" as shell are not returned!
  my $domain = shift @_;
  my @users;

  while (my ($user,$uid,$shell) = (getpwent())[0,2,8]) {
    next if $uid == 0;                       # Skip root/toor
    next if $shell =~ /(?:nologin|uucico)$/; # Skip system users
    push @users, "$user$domain";
  }
  return @users;
}

sub getmbox {
  # This rutine takes a user name as input and returns a reference to a hash
  # containing the user's mailbox'es as keys and corresponding imap folders
  # as value. For virtual domains, the domain must be appended.
  my @users = @_;
  my $box;

  foreach my $user (@users) {
    my ($user, $domain) = $user =~ /^([^\@]+)(\@.*|)/;

    if (defined(scalar getpwnam($user))) { # REAL user
      my ($home) = (getpwnam($user))[7];

      -d $home or
        (warn "Homedir not found: $home\n", return);

      # Locate the INBOX.
      if (-f "$home/Mailbox") {       # INBOX in homedir
        $box->{"$home/Mailbox"} = "user.$user$domain";
      }
      elsif ( -f "/var/mail/$user") { # INBOX in maildir
        $box->{"/var/mail/$user"} = "user.$user$domain";
      }
      else {                          # NO INBOX!
        warn "No main mailbox for $user$domain\n";
        next;
      };

      # Squirrelmail by default stores other mail here (wu-imap)
      foreach my $squirrel (glob("$home/INBOX.*")) {
        next if -z $squirrel;  # Skip empty boxes
        $squirrel =~ /$home\/INBOX\.(.*)/;
        $box->{"$squirrel"} = "user.$user.$1$domain";
      }

      # PINE by default stores other mail here
      foreach my $pine (glob("$home/mail/*")) {
        my $base = basename($pine);   # folder name
        next if $base =~ /^\./;       # Skip hidden files/directories
        next if $base =~ /\.msf$/;    # Mozilla local folder special file

        $box->{"$pine"} = "user.$user.$base$domain";
      }
    }
    elsif (-f "/var/mail/$user" and ! -z  "/var/mail/$user") {
      # Collection list, not a user
      $box->{"/var/mail/$user"} = "user.$user$domain";
    }
  }
  return $box;
}

sub mbox2imap {
  # This rutine takes an Mail::IMAPClient object, unix mailbox filename and
  # imap destination folder as input and extracts mail from the mailbox file
  # into the imap folder. Returns the number of mails succesfully transfered

  my ($client,$mbox,$folder) = @_;
  my $mail;                        # Mail: header and body
  my ($d,$f,$m,$s,$t) = (0,0,0,0,0); # duble, failed, msg no, skip, transfered
  my ($b) = 0;
  my (@new, @replied);

  if (!$client->exists($folder)) {
    print "Creating folder $folder: ";
    $client->create($folder) or
      (print "Could not create $folder: $@\n" and return ($b,$f,$m,$t));
    print "OK\n";
  };

  # Return if mailbox is really a directory or empty
  return ($b,$f,$m,$t) if (-d $mbox or -z $mbox);

  # Select the destination folder. NOTE: There is no unselect, so failing
  # here mess up things, unless error caught!
  $client->select($folder) or
    (print "Could not select: $@\n" and return ($b,$f,$m,$t));

  { # The new line delimiter is set to the mailbox' mail separator (almost).
    # NOTE: The first line of every message will contain some trailing junk
    # to be removed.
    my $tmpbox = $tmpdir . "/" . basename($mbox);
    print "Copying $mbox to $tmpbox: ";
    copy($mbox,$tmpbox) or
      (print "Failed\n" and return ($b,$f,$m,$t));
    print "OK\n";

    local $/ = "\n\nFrom ";
    open(MAIL, "<", $tmpbox) or
      (print "Failed opening $tmpbox: $!\n" and return ($b,$f,$m,$t));

    while (my $msg = <MAIL>) {
      $m++;

      chomp($msg);                   # This removes trailing "\n\nFrom "
      $msg = "X-Envelope: " . $msg;  # Sets non standard header line

      if (!defined($msg) or $msg =~ /^[\n\s]*$/go) {
        print "$mbox: Message $m is empty (no header nor body)\n";
        $s++;
        next;
      };

      my ($header,$body) = split /\n\n/, $msg, 2;

      if (!defined($header) or $header =~ /^[\n\s]*$/go) { # Empty header
        print "$mbox: Message $m has no header.\n";
        print $msg if $debug;
        $s++;
        next;
      }

      if (!defined($body) or $body =~ /^[\n\s]*$/go) {     # Empty body
        $msg .= "\n.\n";                                   # Add a dot
      }

      $header =~ s/\n\s+//go;                              # Unwrap header
      if ($header !~ /^[^\s]+:/gmo) {                      # Invalid header
        print "$mbox: Message $m has invalid header.\n";
        print "$header\n" if $debug;
        $s++;
        next;
      };

      if ($header =~ /^X-IMAP:\s*\d+\s*\d+/igmo) {
        print $header if $debug;
        $s++, next;                                        # Mbox system mail
      }

      my ($msg_id) = $header =~ /^Message-id:\s*([^\n]*)/igmo;
      if (!$msg_id) {
        print "$mbox: Message $m has no message id\n";
        print "$header\n" if $debug;
      }

      # Check if message has been transfered before:
      if ($msg_id and my $id = $client->search(HEADER => 'Message-id' =>
                                   $client->Quote($msg_id))) {
        $d++;
        next;
      };

      $msg =~ s/\n\n> From /\n\nFrom /go;

#      print "Transfering message: ";
      if (my $uid = $client->append($folder,$msg)) {
#        print "$uid, OK\n";
        push @new, $uid;
        $b += do { use bytes; length($msg); };
        if ($header =~ /^X-Status: A/miog) {
          push @replied, $uid;
        }
        $t++;
      } else {
        $f++;
        print "$mbox: Message $m failed: $@\n";
        print $msg, "\n";
      };
    };
    close(MAIL);
    unlink($tmpbox);
  };

  print "Transfered $t messages ($b bytes) of $m total, $f failed, $d dublicates, $s skipped.\n";

  # We mark mails as unseen. This is necesary only to force some mailers to
  # retrieve all mails.
  if (@new) {
    print "Unsetting 'seen' flag: ";
    $client->unset_flag('\Seen',@new) ?
      print "OK\n" : print "Failed\n";
  };
  if (@replied) {
    print "Setting 'answered' flag: ";
    $client->set_flag('\Answered',@replied) ?
      print "OK\n" : print "Failed\n";
  };

  return ($b,$f,$m,$t);
};

#####################################################################
#
# Stub documentation

=head1 mbox2imap.pl

=head2 SYNOPSIS

  mbox2imap.pl [ options users ]

=head2 DESCRIPTION

mbox2imap.pl is yet another a script for transfering Berkely style mailboxes
to a (cyrus) imap server. It uses the imap protocol, hence it should be posible
to transfer mail to any imap-server, however the scripts has been developed for
cyrus and only tested with cyrus.

This script is intended to be run by the administrator for migrating all
users standard Unix mailboxes to the users imap folders. It may be run
multiple times, if a message shares message-id with a message allready in the
destination folder it is not transfered. Messages transfered all have the
'Seen' flag removed (this just appeared to work), and if there is a
"X-Status: R" entry in the header, the 'Answered' flag is set.

If the script is run without arguments the script will lookup mailfolders
for each user in common places in the user home directory and in /var/mail
and place them in corresponding paths.

A list of usernames may be specified, in which case only folders belongning
to those users are transfered. With this option, the administrator can transfer
non-user mailboxes located in /var/mail to the corresponding mailfolder on
the imap server.

For each user the script will look for mailfolders in these places:

Inbox: <homedir>/Mailbox, /var/mail/<user>
Squirrelmail: <homedir>/INBOX.* is transfered to the extension.
PINE: <homedir>/mail/* is transfered to the file name

=head2 USAGE

The script takes a number of command line options, and arguments. The
arguments not associated with an option are usernames or system mailfolders.

=head2 OPTIONS

=item -g arguments are fileglobs, not users. Use this if recovering data from
mailfolders dumped with imap2mbox.pl

=item -h print help text

=item -p imap administrator password, if not specified the script will prompt
for a password

=item -q mailbox quota. This is still a bit experimental. Default is 0, no
quota.

=item -s imap server, default is 'localhost'

=item -u imap administrator user. Note, if you use virtual domains, each
domain may have their own administrator. Default is 'cyrus'.

=item -v verbose/debug

=head2 VIRTUAL DOMAINS

The script supports virtual domains if supported by the server. To transfer
all local users mail to a virtualdomain, simply specify the domain with
'@' prepended, eg '@virtualdomain.com', the list of mailboxes.

To transfer individual users mail into specific virtual domains, list each
user with the full mail account address, eg 'user@virtualdomain.com'

=head2 SKIPPED EMAILS

The script may skip some emails that would otherwise cause errors. Enabling
debugging with the -d debugging flag will cause headers of skipped mail to
be printed, along with other debugging information.

The mbox contains one mail for internal purposes, this is the first mail in
the file, and contains an X-IMAP field in the header. This mail is always
skipped. For this reason, you should expect one mail to be skipped in all
files.

=head2 BUGS AND KNOWN PROBLEMS

The script does not handle ".." in paths.

There is no support for recursion into subdirectories.

Mails with empty body has caused problems, a "." is added as body to these
emails (these e-mails are typically mails sent to list-servers containing
commands in the subject, hence no body).

The Mbox format splits mails with the pattern "\n\nFrom ", some mail clients
does not escape this pattern if occuring in the body of the email. In such
cases, only the part before the "\n\nFrom " will be transfered and the
remaining cause an "invalid header" warning.

=head2 DEPENDENCIES

   IMAP::Admin
   Mail::IMAPClient

The first module provides the methods for creating mail folders, the second
the methods for transfering mail.

=head2 COPYRIGHT

Copyright (C) 2004 Erik Norgaard. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

DISCLAIMER:

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ''AS IS''. ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.

=cut
